import mongoengine
import importlib

from bin.config import config
schemas = importlib.import_module(config["schemas"])


def main():
    mongoengine.connect(config["database_name"], replicaset="monitoring_replSet")
    print("%d Instances found" % len(schemas.Instance.objects()))
    print("%d Results found" % len(schemas.Result.objects()))

    for Instance in schemas.Instance.objects():
        print(Instance.to_json())

    for Result in schemas.Result.objects():
        print(Result.to_json())

    mongoengine.connection.disconnect()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import time
import importlib
import numpy as np
import matplotlib.pyplot as plt
from bin.config import config


# PLOTTING HELPERS

def scatter_plot(ax, x_data, y_data, label):
    # Plot the data, set the size (s), color and transparency (alpha) of the points
    ax.scatter(x_data, y_data, s=10, alpha=0.75, label=label)


def grouped_bar_plot(ax, x_data, y_data_list, y_data_names):
    # Total width for all bars at one x location
    total_width = 0.8
    # Width of each individual bar
    ind_width = total_width / len(y_data_list)
    # This centers each cluster of bars about the x tick mark
    alteration = np.arange(-(total_width / 2), total_width / 2, ind_width)

    # Draw bars, one category at a time
    for i in range(0, len(y_data_list)):
        # Move the bar to the right on the x-axis so it doesn't
        # overlap with previously drawn ones
        ax.bar(x_data + alteration[i], y_data_list[i], label=y_data_names[i], width=ind_width)


# PLOTTING FUNCTIONS

def plot_cactus(data, name, show_date=False, yscale_log=True, out_type="pdf"):
    x_label = "Instance #"
    y_label = "Time (s)"
    title = "Cactus: %s" % " vs. ".join(solver for solver in data.keys())

    if show_date:
        title += " (%s)" % time.strftime("%d/%m/%Y")

    # Create the plot object
    fig, ax = plt.subplots()

    if yscale_log:
        ax.set_yscale('log')

    # Label the axes and provide a title
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    for solver, runs in data.items():
        flt = [r[-1] for r in runs if r[1] in ['sat', 'unsat']]
        scatter_plot(ax, list(range(len(flt))), sorted(flt), solver)

    ax.legend()
    fig.savefig("%s/%s" % ("images/", "%s.%s" % (name, out_type)), bbox_inches='tight')
    plt.close(fig)


def plot_times(data, name, average=True, include_overall=False, show_date=False, out_type="pdf"):
    y_label = "Average Time (s)" if average else "Time (s)"
    title = "Times: %s" % " vs. ".join(solver for solver in data.keys())

    if show_date:
        title += " (%s)" % time.strftime("%d/%m/%Y")

    # Create the plot object
    fig, ax = plt.subplots()

    # Label the axes and provide a title
    ax.set_title(title)
    ax.set_ylabel(y_label)

    choices = ["sat", "unsat", "unknown", "error", "overall"]
    choices = choices if include_overall else choices[:-1]
    x_data = list(range(len(choices)))
    y_data_list = []
    solvers = []

    for solver, runs in data.items():
        solvers.append(solver)
        times = time_results(runs)

        if average:
            count = count_results(runs)
            count = [count[0], count[1], count[2], count[4]]  # remove timeouts
            count = count + [sum(count)] if include_overall else count

            for i in range(len(count)):
                times[i] = times[i] / count[i] if count[i] > 0 else 0

        y_data_list.append(times if include_overall else times[:-1])

    grouped_bar_plot(ax, x_data, y_data_list, solvers)
    ax.set_xticklabels(choices)
    ax.set_xticks(list(range(len(choices))))
    ax.legend()
    fig.savefig("%s/%s" % ("images/", "%s.%s" % (name, out_type)), bbox_inches='tight')
    plt.close(fig)

    print_times(average, choices, solvers, y_data_list)


def plot_counts(data, name, show_date=False, out_type="pdf"):
    y_label = "# Occurences"
    title = "Counts: %s" % " vs. ".join(solver for solver in data.keys())

    if show_date:
        title += " (%s)" % time.strftime("%d/%m/%Y")

    # Create the plot object
    fig, ax = plt.subplots()

    # Label the axes and provide a title
    ax.set_title(title)
    ax.set_ylabel(y_label)

    choices = ["sat", "unsat", "unknown", "timeout", "error"]
    x_data = list(range(len(choices)))
    counts = []
    solvers = []

    for solver, runs in data.items():
        solvers.append(solver)
        counts.append(count_results(runs))

    grouped_bar_plot(ax, x_data, counts, solvers)
    ax.set_xticklabels(choices)
    ax.set_xticks(list(range(len(choices))))
    ax.legend()
    fig.savefig("%s/%s" % ("images/", "%s.%s" % (name, out_type)), bbox_inches='tight')
    plt.close(fig)

    print_counts(choices, solvers, counts)


# ANALYSIS AND AGGREGATION

def count_results(runs):
    choices = ["sat", "unsat", "unknown", "timeout", "error"]
    results = [0 for _ in choices]

    for r in runs["Result"]:
        if r == "sat":
            results[0] += 1
        if r == "unsat":
            results[1] += 1
        if r == "unknown":
            results[2] += 1
        if "timeout" in r:
            results[3] += 1
        if r == "error":
            results[4] += 1

    return results


def time_results(runs):
    choices = ["sat", "unsat", "unknown", "error", "overall"]
    results = [0 for _ in choices]

    for r in range(len(runs["Result"])):
        if runs["Result"][r] == "sat":
            results[0] += runs["Time"][r]
        if runs["Result"][r] == "unsat":
            results[1] += runs["Time"][r]
        if runs["Result"][r] == "unknown":
            results[2] += runs["Time"][r]
        if runs["Result"][r] == "error":
            results[3] += runs["Time"][r]
        results[4] += runs["Time"][r]

    return results


def check_consensus(data):
    # ASSUMING IN SAME ORDER!!!
    issues = []
    min_solved = min(len(runs) for solver, runs in data.items())

    for i in range(min_solved):
        votes = {}

        for solver, runs in data.items():
            votes[solver] = runs["Result"][i]
            problem = runs['Instance'][i]

        done = False
        for _, va in votes.items():
            if done:
                break
            for _, vb in votes.items():
                if done:
                    break
                if va != vb and va in ['sat', 'unsat'] and vb in ['sat', 'unsat']:
                    issues.append((problem, votes))
                    done = True
                    break

    print_consensus_issues(issues)


# PRINTING RESULTS

def print_consensus_issues(issues):
    if len(issues) == 0:
        return

    print("\nDisagreements (%d):" % len(issues))
    print("Instance,", ", ".join(solver for solver in issues[0][1].keys()))

    for i in issues:
        print("%s," % i[0], ", ".join(i[1][solver] for solver in i[1].keys()))


def print_counts(choices, solvers, counts):
    print("\nCounts:")
    print("solver,", ", ".join(c for c in choices))

    for i in range(len(counts)):
        print(", ".join(c for c in [solvers[i]] + list(map(str, counts[i]))))


def print_times(average, choices, solvers, times):
    print("\nAverage Times (s):") if average else print("\nTimes (s):")
    print("solver,", ", ".join(c for c in choices))

    for i in range(len(times)):
        print(", ".join(c for c in [solvers[i]] + list(map(repr, times[i]))))


def import_category():
    schemas = importlib.import_module(config["schemas"])
    return schemas.read_database()


# ENTRY POINT

def analyze():
    data = import_category()

    check_consensus(data)
    plot_cactus(data, "overall_cactus")
    plot_counts(data, "overall_counts")
    plot_times(data, "overall_times")



config = {'schemas': 'bin.sat.schemas', 'commands': {'maple': {'maple_stock': 'tools/sat/maple/MapleLRB_LCMoccRestart/sources/simp/glucose_static -cpu-lim=33'}, 'minisat': {'minisat_clone': 'minisat -cpu-lim=33', 'minisat_stock': 'minisat -cpu-lim=33'}, 'syrup': {'syrup_stock': 'tools/sat/syrup/syrup/bin/glucose-syrup -nthreads=8 -model -maxmemory=50000 -cpu-lim=33'}}, 'handlers': {'maple': 'bin.sat.maple_handler', 'minisat': 'bin.sat.minisat_handler', 'syrup': 'bin.sat.syrup_handler'}, 'timeout': 30.0, 'instances': 'bin/sat/instances', 'database_name': 'sat_database'}



#!/usr/bin/env python3
import sys
import json
import glob
import progressbar
import mongoengine
import importlib
from pymongo import MongoClient
from subprocess import Popen, DEVNULL
from multiprocessing import Process


def write_config(config):

    # convert modules into importable format
    config["schemas"] = config["schemas"].rsplit(".", 1)[0].replace("/", ".")
    for program, handler in config["handlers"].items():
        config["handlers"][program] = handler.rsplit(".", 1)[0].replace("/", ".")

    with open("bin/config.py", "w") as file:
        file.write("config = " + str(config) + "\n")
        file.flush()


def write_instances(config, instances):
    schemas = importlib.import_module(config["schemas"])

    mongoengine.connect(config["database_name"], replicaset="monitoring_replSet")

    for instance in instances:
        stripped_instance = instance.split("/", 1)[1]

        if not schemas.Instance.objects(filename=stripped_instance):
            schemas.Instance.objects(filename=stripped_instance).\
                update_one(upsert=True, set__filename=stripped_instance)

    mongoengine.connection.disconnect()


def collect_handlers(config):

    handlers = {}

    for program in config["handlers"].items():
        cur_module = importlib.import_module(program[1])
        handlers[program[0]] = cur_module.output_handler

    return handlers


def monitor_database(config, num_instances, num_bench):
    commands = config["commands"]

    num_commands = 0
    for program in list(commands.values()):
        num_commands += len(list(program.values()))

    print("Running %d total commands\n" % (num_commands * num_instances * num_bench))

    client = MongoClient()
    db = client[config["database_name"]]

    with db.watch([{'$match': {'operationType': 'insert'}}]) as stream:

        for _ in progressbar.progressbar(range(num_commands * num_instances * num_bench)):
            stream.next()
            # print(stream.next()["fullDocument"])  TODO: possibly use for live-updating output


def main():

    config = json.loads(open(sys.argv[1], 'r').read())
    write_config(config)

    Popen("mongod --dbpath ./results --logpath ./results/log/mongodb.log".split() +
          " --replSet monitoring_replSet".split(), stdout=DEVNULL)

    num_bench = 1
    if len(sys.argv) > 2 and int(sys.argv[2]) >= 0:  # running bench 0 times just calls analyze
        num_bench = int(sys.argv[2])

    if num_bench != 0:

        instances = glob.glob("%s/**/*.*" % config["instances"], recursive=True)

        instance_writer = Process(target=write_instances, args=(config, instances))
        instance_writer.start()
        handlers = collect_handlers(config)
        instance_writer.join()

        database_monitor = Process(target=monitor_database, args=(config, len(instances), num_bench))
        database_monitor.start()

        from bin.bench import bench
        for _ in range(0, num_bench):
            bench(instances, handlers)

    from bin.analyze import analyze
    analyze()

    # TODO: find and kill any running mongod process


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os
import sys
import subprocess
import signal
import datetime
import concurrent.futures
from bin.config import config


def run_problem(program, nickname, command, instance):
    # pass the problem to the command
    invocation = "%s %s" % (command, instance)
    # get start time
    start = datetime.datetime.now().timestamp()
    # run command
    process = subprocess.Popen(
        invocation,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        preexec_fn=os.setsid
    )
    # wait for it to complete
    try:
        process.wait(timeout=config["timeout"])
    # if it times out ...
    except subprocess.TimeoutExpired:
        # kill it
        # print('TIMED OUT:', repr(invocation), '... killing', process.pid, file=sys.stderr)
        os.killpg(os.getpgid(process.pid), signal.SIGINT)
        # set timeout result
        elapsed = config["timeout"]
        output = 'timeout (%.1f s)' % config["timeout"]
    # if it completes in time ...
    else:
        # measure run time
        end = datetime.datetime.now().timestamp()
        elapsed = end - start
        # get result
        stdout = process.stdout.read().decode("utf-8", "ignore")
        stderr = process.stderr.read().decode("utf-8", "ignore")
        output = stdout + stderr
    OUTPUT_HANDLERS[program](nickname, instance, output, elapsed)


# program, specification["id"], specification["command"], problems
def run_solver(args):
    program = args[0]
    nickname = args[1]
    command = args[2]
    instances = args[3]

    for instance in instances:
        run_problem(program, nickname, command, instance)


def signal_handler():
    print("KILLING!")
    try:
        sys.exit(0)
    except SystemExit:
        exit(0)


def bench(instances, handlers):
    global OUTPUT_HANDLERS
    OUTPUT_HANDLERS = handlers

    signal.signal(signal.SIGTERM, signal_handler)

    args = [[program, nickname, command, instances] for
            program, specifications in config["commands"].items() for
            nickname, command in specifications.items()]
    try:
        with concurrent.futures.ProcessPoolExecutor() as executor:
            executor.map(run_solver, args)
    except KeyboardInterrupt:
        print('Interrupted!')
        try:
            sys.exit(0)
        except SystemExit:
            exit(0)



#!/usr/bin/env python3
import os
import psutil
import setuptools
import platform
import time
from subprocess import run, Popen, DEVNULL


print("Creating directory structure")
os.makedirs("results/log", exist_ok=True)
os.makedirs("images", exist_ok=True)

uid = os.getuid()
os.chown("results", uid, -1)
os.chown("results/log", uid, -1)
os.chown("images", uid, -1)


print("Calling correct OS MongoDB install script")
operating_system = platform.system()
if operating_system == "Linux":
    run(['./bin/install_mongodb/linux.sh'])
else:
    print("OS not currently supported :(")
    exit(1)


print("Creating new database and initiate replica set")
Popen("mongod --dbpath ./results --logpath ./results/log/mongodb.log".split() +
      " --replSet monitoring_replSet".split(), stdout=DEVNULL)

code = 1
while not code == 0:
    code = run("mongo --eval 'rs.initiate()'".split(), stdout=DEVNULL, stderr=DEVNULL).returncode
    time.sleep(.5)


PROCNAME = "mongod"

for proc in psutil.process_iter():
    if proc.name() == PROCNAME:
        proc.kill()


print("Calling setuptools.setup function")
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiutare",
    version="1.0",
    author="Federico Mora, Lukas Finnbarr O'Callahan",
    author_email="fmora@cs.toronto.edu, lukasocallahan@gmail.com",
    description="A benchmarking framework for SAT, SMT, and equivalence checking programs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FedericoAureliano/aiutare",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=['mongoengine', 'matplotlib', 'numpy', 'progressbar2', 'pymongo', 'psutil']
)
